/**************************************************************************
* iptrk - ip tracker
* Copyright (C) 2003 Jason Nunn <jsno@leal.com.au>
* Darwin, NT, Australia.
* Originally developed for Arafura Connect (Dec 2000 - Mar 2003)
* Donated to GPL before liquidation.
*
* =======================================================================
* maintenance code (logging)
*
**************************************************************************/
#include <curses.h>
#include "defs.h"
#include "lib.p"
#include "iptrk.p"

typedef struct
{
  port_key_t port;
  voldr_t    voldr;
} sort_port_t;   

PRIVATE FILE    *fp;
PRIVATE addr_t  kaddr, *kaddr_p;
PRIVATE voldr_t log_total;
PRIVATE treap_t *sort_port_oth_bt,
                *sort_port_tcp_bt,
                *sort_port_udp_bt,
                *sort_port_oth_total_bt,
                *sort_port_tcp_total_bt,
                *sort_port_udp_total_bt;

/**************************************************************************
* sort PORT update routines
*
* used to produce nice sorted port listinga for loggs & reporting
**************************************************************************/
PRIVATE void update_sort_port_entry(treap_t **bt, sort_port_t *entry)
{
  sort_port_t *p;

  int find_cmp_func(void *n)
  {
    return cmp_32(entry->port.key, ((sort_port_t *)n)->port.key);
  }

  int append_cmp_func(void *n1, void *n2)
  {
    return cmp_32(((sort_port_t *)n1)->port.key, ((sort_port_t *)n2)->port.key);
  }

  p = treap_find(bt, find_cmp_func);
  if(p == NULL)
  {
    treap_append(bt, entry, sizeof(sort_port_t), append_cmp_func);

    p = treap_find(bt, find_cmp_func);
    if(p == NULL)
    {
      log_m('E', __func__, "::treap_find(): failed");
      exit(-1);
    }
  }
  else
    voldr_acc(&(p->voldr), &(entry->voldr));
}

/**************************************************************************
*
**************************************************************************/
PRIVATE inline void print_port(U_16 p)
{
  if(!p)
    fprintf(fp, "    *");
  else
    fprintf(fp, "%5hu", p);
}

PRIVATE void print_sort_port(treap_t *sort_port, char *title)
{
  char str1[STR_LEN + 1],
       str2[STR_LEN + 1],
       str3[STR_LEN + 1],
       str4[STR_LEN + 1],
       str5[STR_LEN + 1];

  void w_tree(treap_t *t)
  {
    sort_port_t *n;

    if(t == NULL)
      return;

    w_tree(t->l);

    n = TREAP_ENTRY(t);
    print_port(n->port.port.s);
    fprintf(fp, " -> ");
    print_port(n->port.port.d);

    print_vol_u64(str1, n->voldr.I);
    print_vol_u64(str2, n->voldr.O);
    print_vol_u64(str3, n->voldr.u);
    print_vol_u64(str4, n->voldr.i);
    print_vol_u64(str5, n->voldr.o);
    fprintf(fp, ": %20s %20s %20s %20s %20s\n",
      str1, str2, str3, str4, str5);

    w_tree(t->r);
  }

  fprintf(fp, "%s:\n", title);
  w_tree(sort_port);
  putc('\n', fp);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void calc_entry(iptrk_t *i)
{
  int         dir;
  sort_port_t tmp_sp;
  treap_t     **sortp_bt;

  void walk_tree_append_sortp(treap_t *t)
  {
    port_t *n;

    if(t == NULL)
      return;

    walk_tree_append_sortp(t->l);

    n = TREAP_ENTRY(t);
    tmp_sp.port.key = n->port.key;
    cast_volume_log_sub(
      i->addr.addr.s, i->addr.addr.d, kaddr_p, n->vol,
      &(tmp_sp.voldr), dir);
    update_sort_port_entry(sortp_bt, &tmp_sp);

    walk_tree_append_sortp(t->r);
  }

  dir = cast_volume_log(
    i->addr.addr.s, i->addr.addr.d, kaddr_p, i->vol, &(tmp_sp.voldr));
  voldr_acc(&log_total, &(tmp_sp.voldr));

  sortp_bt = &sort_port_oth_bt;
  walk_tree_append_sortp(i->oth_port_ll);

  sortp_bt = &sort_port_tcp_bt;
  walk_tree_append_sortp(i->tcp_port_ll);

  sortp_bt = &sort_port_udp_bt;
  walk_tree_append_sortp(i->udp_port_ll);

/*do totals*/
  sortp_bt = &sort_port_oth_total_bt;
  walk_tree_append_sortp(i->oth_port_ll);

  sortp_bt = &sort_port_tcp_total_bt;
  walk_tree_append_sortp(i->tcp_port_ll);

  sortp_bt = &sort_port_udp_total_bt;
  walk_tree_append_sortp(i->udp_port_ll);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void search_tree(treap_t *t)
{
  iptrk_t *n;

  if(t == NULL)
    return;

  search_tree(t->l);

  n = TREAP_ENTRY(t);
  if(kaddr_p->addr == (n->addr.addr.s & kaddr_p->mask))
    calc_entry(n);
  else
    if(kaddr_p->addr == (n->addr.addr.d & kaddr_p->mask))
      calc_entry(n);

  search_tree(t->r);
}

/**************************************************************************
* log an individual IP address
**************************************************************************/
PRIVATE void log_addr(void)
{
  char           str[STR_LEN + 1],
                 str1[STR_LEN + 1],
                 str2[STR_LEN + 1],
                 str3[STR_LEN + 1],
                 str4[STR_LEN + 1],
                 str5[STR_LEN + 1],
                 filename[STR_LEN + 1];
  struct in_addr in;

  memset(&log_total, 0, sizeof(voldr_t));
  search_tree(watch_bt);

  in.s_addr = htonl(kaddr_p->addr);
  sprintf(filename, IPTRK_LOG_IP_PATHC "/log.%s", inet_ntoa(in));
  fp = fopen(filename, "a");
  if(fp != NULL)
  {
    print_vol_u64(str1, log_total.I);
    print_vol_u64(str2, log_total.O);
    print_vol_u64(str3, log_total.u);
    print_vol_u64(str4, log_total.i);
    print_vol_u64(str5, log_total.o);
    sprintf(str, "\n\t\t%20s %20s %20s %20s %20s\n",
      str1, str2, str3, str4, str5);

    fprintf(fp, "o ");
    print_log_sub(str1, str);
    fputs(str1, fp);

    print_sort_port(sort_port_tcp_bt, "TCP");
    print_sort_port(sort_port_udp_bt, "UDP");
    print_sort_port(sort_port_oth_bt, "OTH");

    fclose(fp);
  }

  treap_free(&sort_port_oth_bt);
  treap_free(&sort_port_tcp_bt);
  treap_free(&sort_port_udp_bt);
}

/**************************************************************************
* log ports
**************************************************************************/
PRIVATE void log_ports(void)
{
  char str[STR_LEN + 1];

  fp = fopen(IPTRK_LOG_PORT_FILEC, "a");
  if(fp != NULL)
  {
    fprintf(fp, "o ");
    print_log_sub(str, "\n");
    fputs(str, fp);
    print_sort_port(sort_port_tcp_total_bt, "TCP");
    print_sort_port(sort_port_udp_total_bt, "UDP");
    print_sort_port(sort_port_oth_total_bt, "OTH");
    fclose(fp);
  }

  treap_free(&sort_port_oth_total_bt);
  treap_free(&sort_port_tcp_total_bt);
  treap_free(&sort_port_udp_total_bt);
}

/**************************************************************************
*
**************************************************************************/
PRIVATE void walk_local_addrs(treap_t *t)
{
  if(t == NULL)
    return;

  walk_local_addrs(t->l);

  kaddr_p->addr = *((U_32 *)TREAP_ENTRY(t));
  log_addr();

  walk_local_addrs(t->r);
}

/**************************************************************************
* LOG
**************************************************************************/
PUBLIC void do_log(void)
{
  ll_t *ll;

  log_m('I', NULL, " doing logging");

  kaddr_p = &kaddr;
  kaddr_p->mask = 0xffffffff;
  walk_local_addrs(our_net_ip_bt);

  ll = boot.extra_report_ll;
  for(;;)
  {
    kaddr_p = ll_read_entry(&ll);
    if(kaddr_p == NULL)
      break;

    log_addr();
  }

  log_ports();
}
